'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _resourceRouterMiddleware = require('resource-router-middleware');

var _resourceRouterMiddleware2 = _interopRequireDefault(_resourceRouterMiddleware);

var _util = require('../lib/util');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 * Retourne les tags à valider par l'utilisateur pour un object
 * @param {*} db 
 * @param {*} user 
 * @param {*} object
 * @param {*} footprint
 */
function getAuthorizedTags(db, user, object, footprint) {
	var Theme = db.theme;
	var ObjectGroup = db.objectGroup;
	var Right = db.right;
	var BusinessGroup = db.businessGroup;
	var User = db.user;
	var Op = db.Sequelize.Op;

	return new Promise(function (resolve, reject) {

		Theme.findAll({
			where: {
				'$objects.rights.business.users.u_id$': _defineProperty({}, Op.eq, user.id),
				// Ne s'occupe pas des groupes d'objets inactifs
				'$objects.og_active$': _defineProperty({}, Op.eq, true)
			},
			order: [['name', 'ASC'], [Theme.ObjectGroup, 'name', 'ASC']],
			include: [{
				model: ObjectGroup,
				as: 'objects',
				include: [{
					association: ObjectGroup.Keys
				}, {
					model: Right,
					as: 'rights',
					include: [{
						association: Right.Keys
					}, {
						model: BusinessGroup,
						as: 'business',
						include: [{
							model: User,
							as: 'users',
							// only attribute for restrict results without loading User
							attributes: ['u_id']
						}]
					}]
				}]
			}]
		}).then(function (themes) {
			(0, _util.getThemesOsmDiff)(db, themes, footprint, null).then(function (tree) {
				var authorizedKeys = [];
				for (var i = 0; i < tree.length; i++) {
					var themetags = getAuthorizedTagsByObject_(tree[i], object);
					if (themetags) {
						authorizedKeys = [].concat(_toConsumableArray(authorizedKeys), _toConsumableArray(themetags));
					}
				}
				resolve(authorizedKeys);
			});
		});
	});
}

/**
 * Retourne tous les tags à valider pour un object (tous les utilisateurs)
 * @param {*} db 
 * @param {*} object
 * @param {*} footprint
 */
function getTagsToValidate(db, object, footprint) {
	var Theme = db.theme;
	var ObjectGroup = db.objectGroup;
	var Right = db.right;
	var BusinessGroup = db.businessGroup;
	var Op = db.Sequelize.Op;

	return new Promise(function (resolve, reject) {

		Theme.findAll({
			order: [['name', 'ASC'], [Theme.ObjectGroup, 'name', 'ASC']],
			include: [{
				model: ObjectGroup,
				as: 'objects',
				include: [{
					association: ObjectGroup.Keys
				}, {
					model: Right,
					as: 'rights',
					include: [{
						association: Right.Keys
					}, {
						model: BusinessGroup,
						as: 'business'
					}]
				}]
			}]
		}).then(function (themes) {
			(0, _util.getThemesOsmDiff)(db, themes, footprint, null).then(function (tree) {
				var authorizedKeys = [];
				for (var i = 0; i < tree.length; i++) {
					var themetags = getAuthorizedTagsByObject_(tree[i], object);
					if (themetags) {
						authorizedKeys = [].concat(_toConsumableArray(authorizedKeys), _toConsumableArray(themetags));
					}
				}
				resolve(authorizedKeys);
			});
		});
	});
}

/**
 * Retourne le groupe d'objets en fonction d'un osmdiff id
 * @param {*} tree
 * @param {*} object 
 */
function getGroupByObject_(tree, object) {
	if (tree.objects) {
		for (var ii = 0; ii < tree.objects.length; ii++) {
			var group = tree.objects[ii];
			if (object.og_gr_id === group.id) {
				return group;
			}
		}
	}
}

/**
 * Retourne les tags autorisés pour un osmdiff id
 * @param {*} themes : thème associé aux groupes d'objets qui contiennent les droits et les objets
 * @param {*} object : objet OSM en cours de validation
 */
function getAuthorizedTagsByObject_(themes, object) {
	// group = tous les objets du groupe d'objet
	// group.rights = droits du groupe d'objet
	var group = getGroupByObject_(themes, object);

	var authorizedTags = [];
	if (group && group.rights) {
		for (var i = 0; i < group.rights.length; i++) {
			var right = group.rights[i];

			// Si tous les tags sont à vérifier, on ajoute tous les tags comme tags validés dans la table lbt_validation
			if (right.r_verif_all_tags) {
				var keys = Object.keys(object.added_tags).concat(Object.keys(object.deleted_tags)).concat(Object.keys(object.modified_tags));
				for (var ii = 0; ii < keys.length; ii++) {
					var key = keys[ii];
					authorizedTags.push(key);
				}
			}
			// Sinon on ajoute uniquement les tags à valider issus des droits
			else if (right.keys) {
					for (var _ii = 0; _ii < right.keys.length; _ii++) {
						var rightKey = right.keys[_ii];
						if (rightKey.name) {
							authorizedTags.push(rightKey.name);
						}
					}
				}

			// Si la géométrie est à vérifier, on ajoute un faux tag "lbt_geom" dans la table lbt_validation
			if (right.r_verif_geom) {
				authorizedTags.push('lbt_geom');
			}
		}
	} else {
		return null;
	}

	//return authorizedTags;
	// Supprime les éventuels tags en doublon (si plusieurs droits)
	return authorizedTags.filter(function (value, index) {
		return authorizedTags.indexOf(value) === index;
	});
}

exports.default = function (_ref) {
	var config = _ref.config,
	    db = _ref.db;
	return (0, _resourceRouterMiddleware2.default)({

		/** Property name to store preloaded entity on `request`. */
		id: 'validation',

		/** For requests with an `id`, you can auto-load the entity.
   *  Errors terminate the request, success sets `req[id] = data`.
   */
		load: function load(req, id, callback) {
			var Validation = db.validation;

			Validation.findByPk(id, {
				include: [Validation.User, Validation.Status]
			}).then(function (validation) {
				return callback(null, validation);
			}).catch(function (err) {
				return callback(console.log(err));
			});
		},


		/** GET / - List all entities */
		list: function list(_ref2, res) {
			var params = _ref2.params;

			var Validation = db.validation;

			Validation.findAll({
				include: [Validation.User, Validation.Status]
			}).then(function (validations) {
				return res.json(validations);
			}).catch(function (err) {
				return res.status(500).json({
					error: err
				});
			});
		},


		/** POST / - Create a new entity */
		create: function create(_ref3, res) {
			var body = _ref3.body;


			var Validation = db.validation;
			var OsmDiff = db.osmdiff;
			var Op = db.Sequelize.Op;

			body.val_st_id = body.status.id;
			body.val_u_id = body.user.id;
			body.val_od_id = body.object.id;

			// const footprint = body.footprint;
			var footprint = 'none';

			// Tous les tags à valider qu'importe l'utilisateur (contiendra lbt_geom si présent)
			getTagsToValidate(db, body.object, footprint).then(function (objectAllTags) {
				console.log('object : all tags to validate (global)', objectAllTags);
				body.allTags = objectAllTags;

				// Tous les tags à valider par l'user en cours
				getAuthorizedTags(db, body.user, body.object, footprint).then(function (objectTags) {
					console.log('object : all tags to validate by user', objectTags);
					body.tags = objectTags;

					// Tous les tags modifiés sur l'objet en cours
					body.allModifiedTags = [].concat(Object.keys(body.object.deleted_tags), Object.keys(body.object.modified_tags), Object.keys(body.object.added_tags));
					// Si la géométrie de l'objet a été modifiée, on y ajoute le tag "lbt_geom"
					if (body.object.changed_geom) {
						body.allModifiedTags.push('lbt_geom');
					}
					console.log('object : all modified tags for this object', body.allModifiedTags);

					// Création de la validation
					Validation.create(body).then(function (created) {

						// Si status vaut invalid ou waiting, ne vérifie pas la liste des tags
						if (['invalid', 'waiting'].includes(body.status.name)) {
							// don't check validation tags list
							return OsmDiff.findByPk(body.object.id).then(function (osm) {
								return osm.setStatus(body.status.id);
							}).then(function () {
								return created;
							});
						}

						// Status vaut valid
						// Recherche dans les validations avec tags non null qui ont été faites sur le même objet
						return Validation.findAll({
							where: {
								tags: _defineProperty({}, Op.ne, null)
							},
							include: [{
								association: Validation.Object,
								attributes: ['id'],
								where: {
									id: body.object.id
								}
							}]
						}).then(function (validations) {

							// Tags des validations sur le même objet (nouvellement et anciennement validées)
							var validedTags = validations.reduce(function (a, v) {
								var _ref4;

								return (_ref4 = []).concat.apply(_ref4, [a].concat(_toConsumableArray(v.tags)));
							}, []);
							console.log('validedTags', validedTags);

							// Si aucun tag validé
							if (!validedTags.length) {
								// Change "status" de l'osmdiff en created
								return OsmDiff.findByPk(body.object.id).then(function (osm) {
									if (osm.changed_geom) {
										return osm.setStatus(body.status.id).then(function () {
											return created;
										});
									}
									console.error('no validation tags found for object ' + body.object.id);
									return created;
								});
							}

							// Tags restants à valider théoriquement (tous les tags à valider - tous les tags validés)
							var tagsToValidate = body.allTags.filter(function (x) {
								return !validedTags.includes(x);
							});

							// Tags restants à valider réellement (intersection entre tags réellement présents dans l'objet et tags restants à valider théoriquement)
							var tagsRemainToValidate = body.allModifiedTags.filter(function (x) {
								return tagsToValidate.includes(x);
							});
							console.log('tags remain', tagsRemainToValidate);

							// Si des tags restent à valider
							if (tagsRemainToValidate.length) {
								return created;
							} else {
								// Autrement set le status de l'osmdiff à created (objet validé)
								return OsmDiff.findByPk(body.object.id).then(function (osm) {
									return osm.setStatus(body.status.id);
								}).then(function () {
									return created;
								});
							}
						});
					}).then(function (validation) {
						return Validation.findByPk(validation.id, {
							include: [Validation.User, Validation.Status, Validation.Object]
						});
					}).then(function (reloaded) {
						return res.json(reloaded);
					}).catch(function (err) {
						return res.status(500).json({
							error: err
						});
					});
				});
			});
		},


		/** GET /:id - Return a given entity */
		read: function read(_ref5, res) {
			var validation = _ref5.validation;

			res.json(validation);
		},


		/** PUT /:id - Update a given entity */
		update: function update(_ref6, res) {
			var validation = _ref6.validation,
			    body = _ref6.body;

			res.status(501);
		},


		/** DELETE /:id - Delete a given entity */
		delete: function _delete(_, res) {
			res.status(501);
		}
	});
};