"use strict";

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.toRes = toRes;
exports.getSettings = getSettings;
exports.getFootprintGeom = getFootprintGeom;
exports.getThemesOsmDiff = getThemesOsmDiff;
exports.getThemeOsmDiff_ = getThemeOsmDiff_;

var _themeObject = require("./themeObject");

/**	Creates a callback that proxies node callback style arguments to an Express Response object.
 *	@param {express.Response} res	Express HTTP Response
 *	@param {number} [status=200]	Status code to send on success
 *
 *	@example
 *		list(req, res) {
 *			collection.find({}, toRes(res));
 *		}
 */
function toRes(res) {
	var status = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 200;

	return function (err, thing) {
		if (err) return res.status(500).send(err);

		if (thing && typeof thing.toObject === 'function') {
			thing = thing.toObject();
		}
		res.status(status).json(thing);
	};
}

/**
 * Récupère les settings
 * @param {*} db 
 * @returns 
 */
function getSettings(db) {
	return new Promise(function (resolve, reject) {

		var Setting = db.setting;
		Setting.findAll().then(function (res) {
			var settings = res.reduce(function (a, s) {
				a[s.name] = s.value;
				return a;
			}, {});
			resolve(settings);
		}).catch(function (err) {
			reject(err);
		});
	});
}

/**
 * Retourne la géométrie d'un footprint
 * @param {*} db 
 * @param {*} footprintId 
 * @returns 
 */
function getFootprintGeom(db, footprintId) {
	return new Promise(function (resolve, reject) {
		var Footprint = db.footprint;
		Footprint.findByPk(footprintId).then(function (footprint) {

			var insersectGeom = void 0;
			if (footprint && footprint.dataValues && footprint.dataValues.geom) {
				insersectGeom = footprint.dataValues.geom;
			}

			resolve(insersectGeom);
		}).catch(function (err) {
			reject(err);
		});
	});
}

/**
 * Retourne l'arbre des thèmes avec les diffs associées
 * @param {*} db 
 * @param {*} themes 
 * @param {*} footprintId
 * @param {*} filter 
 * @returns 
 */
function getThemesOsmDiff(db, themes, footprintId, filter) {
	return new Promise(function (resolve, reject) {
		// Paramètres
		getSettings(db).then(function (settings) {
			footprintId = footprintId || settings.osm_footprint;

			Promise.all(themes.map(function (theme) {
				return new Promise(function (resolve, reject) {
					getThemeOsmDiff_(db, theme, settings, footprintId, filter).then(function (res) {
						theme = res[0];
						if (theme.objects) {
							var nbThemeDiff = 0;
							for (var i = 0; i < theme.objects.length; i++) {
								if (theme.objects[i].nbOsmDiff) {
									nbThemeDiff += theme.objects[i].nbOsmDiff;
								}
							}
							theme.nbThemeDiff = nbThemeDiff;
						}
						resolve(theme);
					}).catch(function (err) {
						reject(err);
					});
				});
			})).then(function (tree) {
				resolve(tree);
			}).catch(function (err) {
				reject(err);
			});;
		});
	});
}

/**
 * Retourne un thème avec les diffs associées
 * @param {*} db 
 * @param {*} theme 
 * @param {*} settings 
 * @param {*} footprintId 
 * @param {*} filter 
 * @returns 
 */
function getThemeOsmDiff_(db, theme, settings, footprintId, filter) {
	var tmpObjects = [];
	return Promise.all(theme.objects.map(function (object) {
		return new Promise(function (resolve, reject) {
			var themeObjectSrv = new _themeObject.ThemeObjectService(db, object, theme, settings);
			themeObjectSrv.findOsmDiffs(footprintId, filter).then(function (res) {
				var jTheme = theme.toJSON();
				tmpObjects.push(_extends({}, object.toJSON(), {
					osmdiff: res.osmdiff.map(function (d) {

						// Ajout de l'id du groupe
						d.dataValues.og_gr_id = object.id;

						// Retour
						return d.toJSON();
					}),
					nbOsmDiff: res.nbOsmDiff
				}));
				jTheme.objects = tmpObjects;
				resolve(jTheme);
			});
		});
	}));
}